(*
    This file is part of the Collections project -
    which currently provides ternary search trees.
    
    Copyright (C) 2000 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)


(* This is an implemention of a ternary search tree, as
   described in:
   Dr. Dobb's Journal, April 1998, "Ternary Search Trees"
   It's available online at:
   http://www.ddj.com/articles/1998/9804/9804a/9804a.htm
*)
structure TernaryTree :> TERNARY_TREE =
  struct
    (* The datatype for ternary trees containing data *)
    datatype 'a t =
      Leaf of 'a option
    | Node of Char.char * ('a t) * ('a t) * ('a t) * ('a option)

    (* An empty tree *)
    val empty = Leaf NONE

    (* fLookup : 'a t -> (char * substring) option -> 'a option *)
    
    (* The first 2 pattern cases covers when there are no more characters
       in the key string. Then we have found our data - if there is
       any (remember data is of type 'a option) *)
    fun fLookup (Leaf data) NONE = data
      | fLookup (Node (_, _, _, _, data)) NONE = data

      (* End of tree, but more key characters => nothing found *)
      | fLookup (Leaf _) (SOME _) = NONE

      (* More key characters and more tree data => continue search *)
      | fLookup (Node (foundC, lo, e, hi, data)) (s as (SOME(c, cs))) =
          if Char.< (c, foundC) then
            fLookup lo s
          else if Char.> (c, foundC) then
            fLookup hi s
          else
            fLookup e (Substring.getc cs)

    fun lookup table key =
          fLookup table (Substring.getc (Substring.all key))

    (* fInsert : 'a t -> (char * substring) option -> 'a -> 'a t *)

    (* As with fLookup, the first 2 pattern cases convers when there
       are no more characters in the key string. This is where we
       insert the data. Notice that we just overwrite any previous data. *)
    fun fInsert (Leaf _) NONE newData =
          Leaf (SOME(newData))
      | fInsert (Node (foundC, lo, e, hi, _)) NONE newData =
          Node (foundC, lo, e, hi, SOME(newData))

      (* End of tree, but more key characters => insert data in middle subtree *)
      | fInsert (Leaf data) (SOME(c, cs)) newData =
          Node (c, empty,
                   fInsert empty (Substring.getc cs) newData,
                   empty, data)

      (* More key characters and more tree data => continue insert *)
      | fInsert (Node (foundC, lo, e, hi, data)) (s as SOME(c, cs)) newData =
          if Char.< (c, foundC) then
            Node (foundC, fInsert lo s newData, e, hi, data)
          else if Char.> (c, foundC) then
            Node (foundC, lo, e, fInsert hi s newData, data)
          else
            Node (foundC, lo,
                          fInsert e (Substring.getc cs) newData,
                          hi, data)

    fun insert table key data =
          fInsert table (Substring.getc (Substring.all key)) data
  end